#include "tape_format.h"

int BAUD_RATE = DEFAULT_BAUD_RATE,
    BLOCK_SIZE = DEFAULT_BLOCK_SIZE,
    START_BITS = DEFAULT_START_BITS,
    NUMBER_OF_START_BITS = DEFAULT_NUMBER_OF_START_BITS,
    STOP_BITS = DEFAULT_STOP_BITS,
    NUMBER_OF_STOP_BITS = DEFAULT_NUMBER_OF_STOP_BITS;

u_char TAPE_FILENAME[255]; /* length should be <= 10, but anyway, it might be usefull??*/
u_long LOAD_ADDRESS, EXEC_ADDRESS;



u_char WAV_FILE[255];

int USE_WAV_FILE = 0;


/* some signal specific variables. Used in loading */
sigset_t dsp_mask, dsp_oldmask;
sigset_t load_mask, load_oldmask;
int dsp_int = 0,load_int = 0;

int synccount = 0;

/* input audio buffers */
#define AUDIO_BUFFERS 2
u_char *audiobufs[AUDIO_BUFFERS], *audiobuf;
int *audiobufs_length;
int  abuf_index, abuf_number;



/* The crc variable */
u_short crc;
void clear_crc() {
	crc = 0;
}

u_short read_crc() {
	return crc;
}

void byte2crc(u_char b) {
	int i;

	crc ^= b<<8;
	for(i=0;i<8;i++)
		crc = (crc&0x8000) ? ( (crc ^ 0x810) << 1 ) | 1 : crc << 1;
}



void dsp_handler(int dummy) { /* hey, someone is calling dsp child */
	dsp_int = 1;
}

void dsp_pause() {
	sigprocmask(SIG_BLOCK, &dsp_mask, &dsp_oldmask);
	while(!dsp_int)sigsuspend(&dsp_oldmask); /* sleep until new buffer from parent */
	dsp_int = 0;
	sigprocmask(SIG_UNBLOCK, &dsp_mask, NULL);
}

void load_handler(int dummy) { /* hey, someone is calling loader parent */
	load_int = 1;
}

void load_pause() {
	sigprocmask(SIG_BLOCK, &load_mask, &load_oldmask);
	while(!load_int)sigsuspend(&load_oldmask);
	load_int = 0;
	sigprocmask(SIG_UNBLOCK, &load_mask, NULL);
}





int re_sample(int* s, double interval, int n) {
	int i;
	int old_abuf_index = abuf_index;

	for(i=0;i<=n;i++) {
		abuf_index = old_abuf_index + rint((interval*i)/n);

		if (abuf_index >= audiobufs_length[ abuf_number ]) {  /* Buffer shift */

			if (dsp_int && !USE_WAV_FILE)
				fprintf(stderr,"%s: WARNING signal processing too slow.\n", command);
			dsp_pause(); /* await a new workload from parent */
			kill(getppid(), SIGUSR2); /* ok got it */

			old_abuf_index -= audiobufs_length[ abuf_number ];

			if (audiobufs_length[abuf_number] <= 0) {
				fprintf(stderr,"end of audio reached \n");
				return END_OF_AUDIO; /* end of audio */
			}

			abuf_index %= audiobufs_length[ abuf_number ];
			abuf_number = (abuf_number+1) % AUDIO_BUFFERS;
		}
		s[i] = (int)audiobufs[abuf_number][abuf_index];
	}
	return NO_ERROR;
}

int load_byte(u_char *b) {
	static int mask[2], i, bit, idx, status;
	double track = 0;
	int error = 0, s[5], d;

	status = idx = 0;
	(*b) = 0;

	while(status < 3 && !error) {

		for(i=0;i<2;i++) { /* sample the next two half bauds */

			error = re_sample(s,((double)SAMPLE_FREQ)/(2*BAUD_RATE)+track,4);

			mask[i] = ((3*s[1]>s[0]+s[2]+s[4])<<1 | (3*s[3]>s[0]+s[2]+s[4])) ^ ((SIGNAL_PHASE==-1)?0x3:0x0);

			switch (mask[i]) { /* the calculation of track is done in doubles to get nice rounding. */
				/* the value of track has actually a physical meaning. If you really */
				/* want to know then email me.                                       */

			case 0x03:/* 11 -> bit=0 */
				if ((d = s[4]+s[0])>0)
					track = ((double)s[4]-s[0])*SIGNAL_PHASE*SAMPLE_FREQ/(d*8.*BAUD_RATE);
				break;

			case 0x02:/* 10 -> bit=1 */
				if ((d = s[2]+s[4])>0)
					track = ((double)s[2]-s[4])*SIGNAL_PHASE*SAMPLE_FREQ/(d*8.*BAUD_RATE);
				break;

			case 0x01:/* 01 -> bit=1 */
				if ((d = s[4]+s[2])>0)
					track = ((double)s[4]-s[2])*SIGNAL_PHASE*SAMPLE_FREQ/(d*8.*BAUD_RATE);
				break;
	
			case 0x00:/* 00 -> bit=0 */
				if ((d = s[0]+s[4])>0)
					track = ((double)s[0]-s[4])*SIGNAL_PHASE*SAMPLE_FREQ/(d*8.*BAUD_RATE);
				break;
	
			default:
				track = 1; /* Just to avoid getting stuck in a perfect reverse phase */
			}
		}

		switch (mask[0]<<2 | mask[1]) {
		case 0x0c:/* 1100 -> bit=0 */
		case 0x03:
			bit = 0;
			if (synccount < SYNC_LENGTH) synccount = 0;
			break;

		case 0x0a:/* 1010 -> bit=1 */
		case 0x05:
			bit = 1;
			if (synccount < SYNC_LENGTH && ++synccount+1==SYNC_LENGTH) { /* This is also the sync tone */
				fprintf(stderr,"Sync tone detected\n");
				synccount = SYNC_LENGTH;
			}
			break;
	
		case 0x02:/* 0010 -> bit=0 */
		case 0x0d:
			bit = 0;
			if (synccount < SYNC_LENGTH) synccount = 0;
			break;
	
		case 0x0b:/* 1011 -> bit=1 */
		case 0x04:
			bit = 1;
			if (synccount < SYNC_LENGTH) synccount = 0;
			break;
	
		default:
			if (synccount == SYNC_LENGTH) { /* Arghh! all is lost */
				fprintf(stderr,"%s: unknown signal\n", command);
			}
			synccount = 0;  /* no sync */
			bit = 2;   /* Its an X-bit, uhhh */
		}

		if (synccount == SYNC_LENGTH) { /* we are on the right track */
			switch (status) {
			case 0 : /* waiting for startbits */
				if ((bit<<idx) == (START_BITS&(1<<idx))) {
					idx++;
					if (idx == NUMBER_OF_START_BITS) {
						status = 1;idx = 0; /* now read the byte */
					}
				} else
					if (idx>0) {
						fprintf(stderr,"%s: startbit not correct\n", command);
						error = BAD_AUDIO;
					}
				break;

			case 1 : /* loading data bits     */
				(*b) |= (bit<<idx++);
				if (idx == DATA_BITS) {
					byte2crc(*b); /* include the new byte in the crc value */
					status = 2;idx = 0; /* now expect some stopbits */
				}
				break;

			case 2 : /* loading stopbits     */
				if ((bit<<idx) == (STOP_BITS&(1<<idx))) {
					idx++;
					if (idx == NUMBER_OF_STOP_BITS) status = 3; /* The byte is ready for delivery */
				} else {
					fprintf(stderr,"%s: stopbit not correct\n", command);
					error = BAD_AUDIO;
				}
			}
		}
	}
	return error;
}


int load_lsb_long(u_long* ul) {
	int i, error = 0;
	u_char b;
	(*ul) = 0;
	for(i=0;i<4 && !error;i++) {
		error = load_byte(&b);
		(*ul) |= b<<(i*8);
	}
	return error;
}


int load_lsb_short(u_short* us) {
	int i, error = 0;
	u_char b;
	(*us) = 0;
	for(i=0;i<2 && !error;i++) {
		error = load_byte(&b);
		(*us) |= b<<(i*8);
	}
	return error;
}


void load(int audio) {
	int shmid, i;
	pid_t dsp_child;
	u_char *incoming, byte, flags;
	int error;
	u_long extra;
	u_short block_length, block,expected_block, CCRC, CRC;

#if 1
	/* First allocate some shared memory */
	if (abuf_size == 0) abuf_size = 65536;  /* we are reading from file (i hope) */
	if ((shmid = shmget(IPC_PRIVATE,abuf_size*AUDIO_BUFFERS,IPC_CREAT|IPC_EXCL|384))<0) {
		fprintf(stderr,"%s: Unable to allocate shared memory\n", command);
		exit (-1);
	}
	if ((audiobuf = shmat(shmid,0,0))<0) {
		fprintf(stderr,"%s: Unable to attach shared memory to data segment\n", command);
		exit (-1);
	}
	if ((shmid = shmget(IPC_PRIVATE,sizeof(int)*AUDIO_BUFFERS,IPC_CREAT|IPC_EXCL|384))<0) {
		fprintf(stderr,"%s: Unable to allocate shared memory\n", command);
		exit (-1);
	}
	if ((audiobufs_length = (int*)shmat(shmid,0,0))<0) {
		fprintf(stderr,"%s: Unable to attach shared memory to data segment\n", command);
		exit (-1);
	}

	for(i=0;i<AUDIO_BUFFERS;i++) /* audio buffer array */
		audiobufs[i] = audiobuf+i*abuf_size;

	*audiobuf = 0;

	if ((dsp_child = fork())) {             /* i am the parent, i will just load audio to  */
		struct sigaction load_action;   /* shared memory and let my child do the work. */

		sigemptyset(&load_mask); sigaddset(&load_mask, SIGUSR2);
		load_action.sa_handler = load_handler;
		load_action.sa_mask = load_mask;
		load_action.sa_flags = 0;
		sigaction(SIGUSR2, &load_action, NULL);

		/* This does not work for some reason (FreeBSD): */
		//while(!(*audiobuf));    /* do some busy wait until child has set up signalling */

		/* But this does: */
		while(!(*audiobuf)) {
			/* do some busy wait until child has set up signalling */
			printf("waiting\n");
		}

		i = 0;

		while((audiobufs_length[i] = read(audio,audiobufs[i],abuf_size)) > 0) {
			kill(dsp_child, SIGUSR1);    /* Hey dsp child! you have some more work!     */
			load_pause();               /* wait for signal from hard working child     */
			i = (i+1) % AUDIO_BUFFERS;
		}
		kill(dsp_child,SIGUSR1);
		wait(NULL);
	} else {                            /* i am the signal proccessing child           */
		struct sigaction dsp_action;

		sigemptyset(&dsp_mask); sigaddset(&dsp_mask, SIGUSR1);

		dsp_action.sa_handler = dsp_handler;
		dsp_action.sa_mask = dsp_mask;
		dsp_action.sa_flags = 0;
		sigaction(SIGUSR1, &dsp_action, NULL);

		*audiobuf = 1; /* ok! ready, signalling set up. */

		dsp_pause(); /* await the first workload from parent */
		kill(getppid(), SIGUSR2); /* ok got your workload, thanx */

		abuf_index = abuf_number = 0;
#endif



		flags = 0; error = 0; expected_block = 0;

//		while(!(flags&0x80) && !error) {
		while(!error) {

			error = load_byte(&byte); /* sync byte */

			if (!error && block == 0 && byte == SYNC_BUGF) {
				fprintf(stderr,"%s: sync bug fix byte %2x recieved\n", command,byte);
				error = load_byte(&byte); /* sync byte */
			}

			if (!error && byte != SYNC_BYTE) {
				fprintf(stderr,"%s: unknown magic number %2x\n", command, byte);
				error = 1;
			}

			clear_crc(); /* start from a fresh crc value */

			i = 0; byte = 1;
			while(!error && byte > 0) {
				error = load_byte(&byte);
				TAPE_FILENAME[i++] = byte;
			}
			if (!error) fprintf(stderr,"%s ", TAPE_FILENAME);
			if (!error) error = load_lsb_long(&LOAD_ADDRESS);
			if (!error) fprintf(stderr,"%8.8x ", (unsigned int) LOAD_ADDRESS);
			if (!error) error = load_lsb_long(&EXEC_ADDRESS);
			if (!error) fprintf(stderr,"%8.8x ", (unsigned int) EXEC_ADDRESS);
			if (!error) error = load_lsb_short(&block);
			if (!error) fprintf(stderr,"%2.2x ", block);
			if (!error) error = load_lsb_short(&block_length);
			if (!error) fprintf(stderr,"%4.4x ", block_length);
			if (!error) error = load_byte(&flags);
			if (!error) fprintf(stderr,"%2.2x \n", flags);
			if (!error) error = load_lsb_long(&extra);

			CCRC = read_crc();
			if (!error) error = load_lsb_short(&CRC);
			CRC = ((CRC & 0xff00) >> 8) | ((CRC & 0x00ff) << 8);
			if (!error && CRC != CCRC) {
				fprintf(stderr, "%s: CRC check failed for header\n", command);
				error = BAD_AUDIO;
			}

			if (!error && expected_block!=block)
				fprintf(stderr, "%s: WARNING expecting block %2.2x.\n", command, expected_block);

			if (!error) {
				if ((incoming = malloc(block_length)) == NULL) {
					fprintf(stderr, "%s: Unable to allocate data buffer for input\n", command);
					exit(-1);
				}

				clear_crc(); /* start from a fresh crc value */

				if (block_length > 0) {
					/* now load the data */
					for(i=0;i<block_length && !error;i++) {
						error = load_byte(&byte);
						incoming[i] = byte;
					}

					/* and now the CRC stuff */
					CCRC = read_crc();
					if (!error) error = load_lsb_short(&CRC);
					CRC = ((CRC&0xff00) >> 8) | ((CRC&0x00ff) << 8);
					if (!error && CRC != CCRC) {
						fprintf(stderr,"%s: CRC check failed for data block\n", command);
						error = BAD_AUDIO;
					}
				}

				if (!error && expected_block == block) {
					FILE *fp;
					FILE *fp_inf;
					char name[256];
					char inf_name[256];

					strcpy(name, TAPE_FILENAME);
					bbcim_replace_bad_chars(name);

					fp = fopen(name, "ab");
					//if (write(data,incoming,block_length) != block_length) {
					if (fwrite(incoming, 1, block_length, fp) != block_length) {
						fprintf(stderr,"%s: error writing data\n", command);
						exit(1);
					}
					fclose(fp);
					if (block == 0) {
						strcpy(inf_name, name);
						strcat(inf_name, ".inf");
						fp_inf = fopen(inf_name, "ab");
						fprintf(fp, "TAPE %-10s %8lX %8lX %s", TAPE_FILENAME, LOAD_ADDRESS, EXEC_ADDRESS, ((flags & 1) ? " Locked" : ""));
						fclose(fp_inf);
					}
				}

				free(incoming);
			}
			if (!error && expected_block == block) expected_block++; /* just perfect */
			else if (error != END_OF_AUDIO) { /* we just try again */
				error = NO_ERROR;
				synccount = 0;
			}
			if (flags & 0x80) expected_block = 0;
		}
		kill(getppid(),SIGINT);
		shmdt((char*)audiobuf);shmdt((char*)audiobufs_length); /* am i right?? */
	}
}
