/* the save wave table */
u_char *one, *zero;
int length;

void make_wave_table() {
	int i;
	length = SAMPLE_FREQ/BAUD_RATE;
	if ((zero = malloc(length))==NULL || (one = malloc(length))==NULL) {
		fprintf(stderr,"%s: Unable to allocate audio buffer for output\n", command);
		exit(-1);
	}
	for(i=0;i<length;i++) /* zero wave table */
		zero[i] = SIGNAL_ZERO+SIGNAL_AMP*SIGNAL_PHASE*sin((2.0*M_PI*i)/length);
	for(i=0;i<length;i++) /* one wave table  */
		one[i] = zero[(2*i)%length];
}


void delete_wave_table() {
	free(zero);free(one);
}


void make_sync(int audio,int sync_length) { /* create the tone */
	int i;

	for(i=0;i<sync_length;i++) {
		if (write(audio,one,length)!=length) {
			perror(AUDIO_DEV);
			exit(-1);
		}
	}
}


void save_byte(int audio, u_char b) {
	int i;
	u_char *onezero;

	byte2crc(b); /* update the crc value */

	for(i=0;i<NUMBER_OF_START_BITS;i++) { /* first write the start bits */
		onezero = (START_BITS&(1<<i)) ? one : zero;
		if (write(audio,onezero,length)!=length) {
			perror(AUDIO_DEV);
			exit(-1);
		}
	}
	for(i=0;i<DATA_BITS;i++) { /* then write the data */
		onezero = (b&(1<<i)) ? one : zero;
		if (write(audio,onezero,length)!=length) {
			perror(AUDIO_DEV);
			exit(-1);
		}
	}
	for(i=0;i<NUMBER_OF_STOP_BITS;i++) { /* finally write the stop bits */
		onezero = (STOP_BITS&(1<<i)) ? one : zero;
		if (write(audio,onezero,length)!=length) {
			perror(AUDIO_DEV);
			exit(-1);
		}
	}
}


void save_long(int audio,u_long ul) {
	int i;
	for(i=0;i<4;i++) {
		save_byte(audio,ul&0xff);
		ul>>=8;
	}
}


void save_short(int audio,u_short us) {
	int i;
	for(i=0;i<2;i++) {
		save_byte(audio, us&0xff);
		us>>=8;
	}
}


void save(int source, int audio) {
	int i;
	u_char *data,flags;
	u_short block = 0,block_length,CRC;

	if ((data = malloc(BLOCK_SIZE))==NULL) {
		fprintf(stderr,"%s: Unable to allocate data buffer\n", command);
		exit(-1);
	}

	make_wave_table(); /* initialize the sound tables */

	make_sync(audio,HEADER_SPACE); /* create the file header tone */

	while((block_length = read(source,data,BLOCK_SIZE))>0) {

		/* write the header first: */

		save_byte(audio,(u_char)0x2a); /* sync byte */

		clear_crc(); /* refresh the crc */

		for(i=0; TAPE_FILENAME[i]!='\0'; i++)
			save_byte(audio, TAPE_FILENAME[i]);
		save_byte(audio,(u_char)0x00);
		fprintf(stderr,"%s ", TAPE_FILENAME);
		save_long(audio,LOAD_ADDRESS);
		fprintf(stderr,"%8.8x ", (unsigned int) LOAD_ADDRESS);
		save_long(audio,EXEC_ADDRESS);
		fprintf(stderr,"%8.8x ", (unsigned int) EXEC_ADDRESS);
		save_short(audio,block);
		fprintf(stderr,"%2.2x ", block++);
		save_short(audio,block_length);
		fprintf(stderr,"%4.4x ", block_length);
		flags = (block_length != BLOCK_SIZE)<<7;
		save_byte(audio,flags);
		fprintf(stderr,"%2.2x \n",flags);
		save_long(audio, 0);

		CRC = read_crc();
		CRC = ((CRC&0xff00) >> 8) | ((CRC&0x00ff) << 8); /* Why this? is it more safe */
		save_short(audio,CRC);

		clear_crc(); /* refresh the crc */

		/* now flush the data */

		for(i=0;i<block_length;i++)
			save_byte(audio,data[i]);

		/* and write data CRC */

		CRC = read_crc();
		CRC = ((CRC&0xff00)>>8)|((CRC&0x00ff)<<8); /* Why this? is it more safe */
		save_short(audio,CRC);

		make_sync(audio,BLOCK_SPACE); /* block sync tone */

	}
	delete_wave_table(); /* clean up before we leave */
}
